from __future__ import annotations
from pathlib import Path

import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import Session, sessionmaker

from crossrs.db.base import Base
from crossrs.db.models import Metadata, Word, Sentence
from crossrs.diff.tokenizer import tokenize, normalize_token
from crossrs.diff import compute_diff
from crossrs.app.commands.init import extract_tokens, extract_words, process_corpus
from crossrs.app.commands.study.updater import (
    update_sentence, extract_words_from_text, INTERVALS,
)
from crossrs.app.commands.study.chooser import (
    choose_next, choose_target_word,
    get_round_counts, get_queue_target_word_ids,
)
from crossrs.utils.time import get_timestamp


# ─── Fixtures ───────────────────────────────────────────────────────────────────

@pytest.fixture
def db_session():
    """Create an in-memory SQLite database session for testing."""
    engine = create_engine('sqlite:///:memory:')
    Base.metadata.create_all(engine)
    session = sessionmaker(bind=engine)()
    yield session
    session.close()


@pytest.fixture
def corpus_file(tmp_path):
    """Create a temporary corpus file."""
    corpus = tmp_path / 'corpus.txt'
    corpus.write_text(
        'Die Katze sitzt auf der Matte.\n'
        'Der Hund läuft im Park.\n'
        'Die Katze und der Hund spielen zusammen.\n'
        'Das Wetter ist heute sehr schön.\n'
        'Ich gehe morgen in die Schule.\n',
        encoding='utf-8',
    )
    return corpus


def populate_db(session: Session, corpus_file: Path) -> dict[str, int]:
    """Populate a test database from a corpus file. Returns ids_by_word."""
    from crossrs.app.commands.init import process_corpus, add_words, add_sentences
    words_by_sentence, word_frequencies = process_corpus(corpus_file)
    ids_by_word = add_words(session, word_frequencies)
    add_sentences(session, words_by_sentence, ids_by_word)
    session.add(Metadata(id=1, total_rounds=0))
    session.commit()
    return ids_by_word


# ─── Tokenizer Tests ───────────────────────────────────────────────────────────

class TestTokenizer:
    def test_basic_tokenization(self):
        tokens = tokenize('Hello world')
        assert len(tokens) == 2
        assert tokens[0].normalized == 'hello'
        assert tokens[1].normalized == 'world'

    def test_unicode_tokenization(self):
        tokens = tokenize('Die Katze läuft.')
        assert len(tokens) == 3
        assert tokens[0].normalized == 'die'
        assert tokens[1].normalized == 'katze'
        assert tokens[2].normalized == 'läuft'

    def test_punctuation_skipped(self):
        tokens = tokenize('Hello, world! How are you?')
        normalized = [t.normalized for t in tokens]
        assert normalized == ['hello', 'world', 'how', 'are', 'you']

    def test_numbers_included(self):
        tokens = tokenize('I have 3 cats')
        assert len(tokens) == 4
        assert tokens[2].normalized == '3'

    def test_position_tracking(self):
        text = 'Hello world'
        tokens = tokenize(text)
        assert text[tokens[0].start:tokens[0].end] == 'Hello'
        assert text[tokens[1].start:tokens[1].end] == 'world'

    def test_normalize_token(self):
        assert normalize_token('HELLO') == 'hello'
        assert normalize_token('Café') == 'café'

    def test_empty_string(self):
        assert tokenize('') == []

    def test_only_punctuation(self):
        assert tokenize('.,!?;:') == []


# ─── Diff Tests ─────────────────────────────────────────────────────────────────

class TestDiff:
    def test_identical_texts(self):
        segments = compute_diff('Hello world', 'Hello world')
        assert segments == []

    def test_single_word_change(self):
        segments = compute_diff('I have a cat', 'I have a dog')
        assert len(segments) == 1
        assert segments[0].old_text == 'cat'
        assert segments[0].new_text == 'dog'

    def test_insertion(self):
        segments = compute_diff('I have cat', 'I have a cat')
        assert len(segments) == 1
        assert segments[0].old_text is None
        assert segments[0].new_text == 'a'

    def test_deletion(self):
        segments = compute_diff('I have a cat', 'I have cat')
        assert len(segments) == 1
        assert segments[0].old_text == 'a'
        assert segments[0].new_text is None

    def test_case_insensitive_comparison(self):
        # Tokens are compared by normalized (lowercased) form
        segments = compute_diff('Hello World', 'hello world')
        assert segments == []


# ─── Word Extraction Tests ─────────────────────────────────────────────────────

class TestWordExtraction:
    def test_extract_tokens(self):
        tokens = extract_tokens('Die Katze sitzt.')
        assert tokens == ['die', 'katze', 'sitzt']

    def test_extract_words(self):
        words = extract_words(['a', 'b', 'c'])
        assert words == ['a', 'b', 'c']

    def test_extract_words_from_text(self):
        words = extract_words_from_text('I have a cat')
        assert 'i' in words
        assert 'have' in words
        assert 'a' in words
        assert 'cat' in words


# ─── Corpus Processing Tests ───────────────────────────────────────────────────

class TestCorpusProcessing:
    def test_process_corpus(self, corpus_file):
        words_by_sentence, frequencies = process_corpus(corpus_file)
        assert len(words_by_sentence) == 5
        # 'die' appears in multiple sentences
        assert frequencies['die'] >= 2
        # Each sentence has its own set of words
        for words in words_by_sentence.values():
            assert isinstance(words, set)
            assert len(words) > 0

    def test_duplicate_sentences_ignored(self, tmp_path):
        corpus = tmp_path / 'dup.txt'
        corpus.write_text('Hello world.\nHello world.\nGoodbye.\n', encoding='utf-8')
        words_by_sentence, frequencies = process_corpus(corpus)
        assert len(words_by_sentence) == 2  # duplicate ignored


# ─── Database / Init Tests ──────────────────────────────────────────────────────

class TestDatabaseInit:
    def test_words_sorted_by_frequency(self, db_session, corpus_file):
        populate_db(db_session, corpus_file)
        words = db_session.query(Word).order_by(Word.id).all()
        # IDs should be in decreasing frequency order
        for i in range(len(words) - 1):
            assert words[i].occurrences >= words[i + 1].occurrences

    def test_sentence_word_associations(self, db_session, corpus_file):
        populate_db(db_session, corpus_file)
        sentence = db_session.query(Sentence).first()
        assert len(sentence.words) > 0

    def test_initial_state(self, db_session, corpus_file):
        populate_db(db_session, corpus_file)
        for sentence in db_session.query(Sentence).all():
            assert sentence.status == 0
            assert sentence.rounds == 0
            assert sentence.due is None
        for word in db_session.query(Word).all():
            assert word.learnedness == 0


# ─── Updater Tests ──────────────────────────────────────────────────────────────

class TestUpdater:
    def test_first_attempt_correct_schedules_last_stage(self, db_session, corpus_file):
        ids = populate_db(db_session, corpus_file)
        sentence = db_session.query(Sentence).first()
        word = db_session.query(Word).first()
        update_sentence(sentence, is_correct=True, is_first_attempt=True,
                        target_word=word, user_translation='The cat sits on the mat.',
                        session=db_session)
        db_session.flush()
        assert sentence.status == 1  # in queue
        assert sentence.review_stage == len(INTERVALS) - 1
        assert sentence.due is not None
        assert sentence.rounds == 1

    def test_first_attempt_incorrect_enters_queue(self, db_session, corpus_file):
        populate_db(db_session, corpus_file)
        sentence = db_session.query(Sentence).first()
        word = db_session.query(Word).first()
        update_sentence(sentence, is_correct=False, is_first_attempt=True,
                        target_word=word, user_translation='wrong', session=db_session)
        db_session.flush()
        assert sentence.status == 1  # in queue
        assert sentence.review_stage == 0
        assert sentence.due is not None
        assert sentence.rounds == 1

    def test_review_correct_advances_stage(self, db_session, corpus_file):
        populate_db(db_session, corpus_file)
        sentence = db_session.query(Sentence).first()
        word = db_session.query(Word).first()
        # Enter queue
        update_sentence(sentence, is_correct=False, is_first_attempt=True,
                        target_word=word, user_translation='wrong', session=db_session)
        db_session.flush()
        # Correct review at stage 0 -> stage 1
        update_sentence(sentence, is_correct=True, is_first_attempt=False,
                        target_word=None, user_translation='correct', session=db_session)
        db_session.flush()
        assert sentence.review_stage == 1
        assert sentence.status == 1

    def test_review_incorrect_resets_stage(self, db_session, corpus_file):
        populate_db(db_session, corpus_file)
        sentence = db_session.query(Sentence).first()
        word = db_session.query(Word).first()
        # Enter queue and advance to stage 1
        update_sentence(sentence, is_correct=False, is_first_attempt=True,
                        target_word=word, user_translation='wrong', session=db_session)
        update_sentence(sentence, is_correct=True, is_first_attempt=False,
                        target_word=None, user_translation='correct', session=db_session)
        db_session.flush()
        assert sentence.review_stage == 1
        # Incorrect at stage 1 -> back to stage 0
        update_sentence(sentence, is_correct=False, is_first_attempt=False,
                        target_word=None, user_translation='wrong', session=db_session)
        db_session.flush()
        assert sentence.review_stage == 0

    def test_full_review_cycle_marks_learned(self, db_session, corpus_file):
        populate_db(db_session, corpus_file)
        sentence = db_session.query(Sentence).first()
        word = db_session.query(Word).first()
        # Enter queue
        update_sentence(sentence, is_correct=False, is_first_attempt=True,
                        target_word=word, user_translation='wrong', session=db_session)
        # Pass all review stages
        for _ in range(len(INTERVALS)):
            update_sentence(sentence, is_correct=True, is_first_attempt=False,
                            target_word=None, user_translation='The cat sits on the mat.',
                            session=db_session)
        db_session.flush()
        assert sentence.status == 2  # learned

    def test_learnedness_incremented_on_learn(self, db_session, corpus_file):
        populate_db(db_session, corpus_file)
        sentence = db_session.query(Sentence).first()
        word = db_session.query(Word).first()
        # First attempt correct → enters queue at last stage
        update_sentence(sentence, is_correct=True, is_first_attempt=True,
                        target_word=word, user_translation='The cat sits on the mat.',
                        session=db_session)
        db_session.flush()
        # Pass the final review → marked as learned
        update_sentence(sentence, is_correct=True, is_first_attempt=False,
                        target_word=None, user_translation='The cat sits on the mat.',
                        session=db_session)
        db_session.flush()
        # Words from the original sentence should have learnedness > 0
        db_session.expire_all()
        sentence_words = sentence.words
        incremented = [w for w in sentence_words if w.learnedness > 0]
        assert len(incremented) > 0

    def test_intervals(self):
        assert int(INTERVALS[0].total_seconds()) == 72000  # 20 hours
        assert int(INTERVALS[1].total_seconds()) == 590400  # 7 days - 4 hours
        assert int(INTERVALS[2].total_seconds()) == 2577600  # 30 days - 4 hours


# ─── Chooser Tests ──────────────────────────────────────────────────────────────

class TestChooser:
    def test_choose_new_sentence(self, db_session, corpus_file):
        populate_db(db_session, corpus_file)
        sentence, target_word, counts = choose_next(db_session, threshold=3)
        assert sentence is not None
        assert target_word is not None
        assert counts.due == 0
        assert sentence.status == 0  # unseen

    def test_target_word_is_most_frequent(self, db_session, corpus_file):
        populate_db(db_session, corpus_file)
        _, target_word, _ = choose_next(db_session, threshold=3)
        # With all learnedness=0, should pick the most frequent word (id=1)
        assert target_word.id == 1

    def test_due_count_zero_initially(self, db_session, corpus_file):
        populate_db(db_session, corpus_file)
        assert get_round_counts(db_session).due == 0

    def test_due_sentence_prioritized(self, db_session, corpus_file):
        populate_db(db_session, corpus_file)
        # Put a sentence in the queue with past due time
        sentence = db_session.query(Sentence).first()
        sentence.status = 1
        sentence.due = get_timestamp() - 1000
        sentence.target_word_id = 1
        db_session.flush()
        result_sentence, target_word, counts = choose_next(db_session, threshold=3)
        assert result_sentence.id == sentence.id
        assert target_word is None  # review, not new
        assert counts.due == 1

    def test_queue_targets_excluded(self, db_session, corpus_file):
        populate_db(db_session, corpus_file)
        # Put first word in queue
        sentence = db_session.query(Sentence).first()
        sentence.status = 1
        sentence.due = get_timestamp() + 100000  # future
        sentence.target_word_id = 1
        db_session.flush()
        queue_targets = get_queue_target_word_ids(db_session)
        assert 1 in queue_targets
        # Next target should not be word 1
        target = choose_target_word(db_session, threshold=3)
        assert target is not None
        assert target.id != 1

    def test_higher_learnedness_preferred(self, db_session, corpus_file):
        populate_db(db_session, corpus_file)
        # Give word 2 higher learnedness
        word2 = db_session.query(Word).filter(Word.id == 2).one()
        word2.learnedness = 2
        db_session.flush()
        target = choose_target_word(db_session, threshold=3)
        assert target.id == 2  # higher learnedness preferred

    def test_threshold_excludes_learned(self, db_session, corpus_file):
        populate_db(db_session, corpus_file)
        # Set all words to learnedness >= threshold
        db_session.query(Word).update({Word.learnedness: 3})
        db_session.flush()
        target = choose_target_word(db_session, threshold=3)
        assert target is None  # all learned


# ─── Stats Tests ────────────────────────────────────────────────────────────────

class TestStats:
    def test_initial_stats(self, db_session, corpus_file):
        from crossrs.app.commands.stats import compute_word_stats, compute_sentence_stats
        populate_db(db_session, corpus_file)

        word_stats = compute_word_stats(db_session, threshold=3)
        sentence_stats = compute_sentence_stats(db_session, threshold=3)

        assert sentence_stats.learned == 0
        assert sentence_stats.in_queue == 0
        assert sentence_stats.total == 5
        assert sentence_stats.total_rounds == 0

        assert word_stats.learned == 0
        assert word_stats.total > 0

    def test_stats_after_learning(self, db_session, corpus_file):
        from crossrs.app.commands.stats import compute_sentence_stats
        populate_db(db_session, corpus_file)

        # First attempt correct → enters queue at last stage (not immediately learned)
        sentence = db_session.query(Sentence).first()
        word = db_session.query(Word).first()
        update_sentence(sentence, is_correct=True, is_first_attempt=True,
                        target_word=word, user_translation='The cat sits on the mat.',
                        session=db_session)
        db_session.flush()

        sentence_stats = compute_sentence_stats(db_session, threshold=3)
        assert sentence_stats.in_queue == 1
        assert sentence_stats.learned == 0
        assert sentence_stats.total_rounds == 1


# ─── Delete Tests ───────────────────────────────────────────────────────────────

class TestDelete:
    def test_delete_with_force(self, tmp_path):
        db_path = tmp_path / 'test.db'
        db_path.write_text('dummy')
        assert db_path.exists()
        db_path.unlink()
        assert not db_path.exists()



from __future__ import annotations

import time
from typing import TypeVar, Callable

from openai import APIError, APIConnectionError, AuthenticationError, RateLimitError

from crossrs.utils.typer import typer_raise

__all__ = [
    'api_call_with_retries',
]

T = TypeVar('T')

MAX_RETRIES = 3
RETRY_DELAYS = [5, 15, 30]


def api_call_with_retries(fn: Callable[[], T], on_error: Callable[[str], None] | None = None) -> T:
    """Call fn() with automatic retries on transient OpenAI errors.

    on_error is called with a message before each retry (e.g. to print to console).
    Raises on non-transient errors or after exhausting retries.
    """
    for attempt in range(MAX_RETRIES + 1):
        try:
            return fn()
        except AuthenticationError:
            typer_raise('Invalid API key. Check --api-key or CROSSRS_API_KEY.')
        except RateLimitError as e:
            _handle_retry(attempt, f'Rate limited: {e}', on_error)
        except APIConnectionError as e:
            _handle_retry(attempt, f'Connection error: {e}', on_error)
        except APIError as e:
            if e.status_code and e.status_code >= 500:
                _handle_retry(attempt, f'Server error: {e}', on_error)
            else:
                typer_raise(f'OpenAI API error: {e.message}')


def _handle_retry(attempt: int, message: str, on_error: Callable[[str], None] | None) -> None:
    if attempt >= MAX_RETRIES:
        raise RuntimeError(f'OpenAI API failed after {MAX_RETRIES + 1} attempts: {message}')
    delay = RETRY_DELAYS[attempt]
    full_message = f'{message} — retrying in {delay}s...'
    if on_error:
        on_error(full_message)
    time.sleep(delay)



from __future__ import annotations
from collections import Counter
from pathlib import Path
from typing import Annotated, Generator

from sqlalchemy import text
from tqdm import tqdm
from typer import Argument

from crossrs.app.app import app
from crossrs.db import get_session, Session
from crossrs.db.models import Metadata, Word, Sentence, SentenceWord
from crossrs.diff.tokenizer import tokenize
from crossrs.utils.typer import typer_raise

__all__ = [
    'init',
]


def get_sentences(corpus: Path) -> Generator[str, None, None]:
    """Yield sentences from the corpus file."""
    with corpus.open('r', encoding='utf-8') as file:
        for line in file:
            if sentence := line.strip():
                yield sentence


def extract_tokens(sentence: str) -> list[str]:
    """Extract normalized tokens from a sentence."""
    return [token.normalized for token in tokenize(sentence)]


def extract_words(tokens: list[str]) -> list[str]:
    """Extract words from a list of tokens."""
    return list(tokens)


def process_corpus(corpus: Path) -> tuple[dict[str, set[str]], Counter[str]]:
    """Process the corpus and return unique words by sentence and word frequencies."""
    words_by_sentence: dict[str, set[str]] = {}
    word_frequencies: Counter[str] = Counter()

    for sentence in tqdm(get_sentences(corpus), desc='Processing sentences'):
        if sentence not in words_by_sentence:
            tokens = extract_tokens(sentence)
            if tokens:
                words = extract_words(tokens)
                words_by_sentence[sentence] = set(words)
                word_frequencies.update(words)

    return words_by_sentence, word_frequencies


def add_words(session: Session, frequencies: Counter[str]) -> dict[str, int]:
    """Add words to the database sorted by frequency and return a mapping of words to their IDs."""
    ids_by_word: dict[str, int] = {}
    for word, occurrences in tqdm(frequencies.most_common(), desc='Adding words'):
        word_obj = Word(
            word=word,
            occurrences=occurrences,
            learnedness=0,
        )
        session.add(word_obj)
        session.flush()
        ids_by_word[word] = word_obj.id
    return ids_by_word


def add_sentences(session: Session, words_by_sentence: dict[str, set[str]],
                  ids_by_word: dict[str, int]) -> None:
    """Add sentences to the database with their word associations."""
    for sentence_text, words in tqdm(words_by_sentence.items(), desc='Adding sentences'):
        sentence_obj = Sentence(sentence=sentence_text)
        session.add(sentence_obj)
        session.flush()
        for word in words:
            session.add(SentenceWord(
                sentence_id=sentence_obj.id,
                word_id=ids_by_word[word],
            ))
        session.flush()


@app.command()
def init(
        language: Annotated[
            str,
            Argument(help='Target language code (e.g., "de", "fr", "uk").'),
        ],
        corpus: Annotated[
            Path,
            Argument(
                dir_okay=False,
                exists=True,
                readable=True,
                resolve_path=True,
                help='Plain-text file containing one sentence per line.',
            ),
        ],
) -> None:
    """Initialize CrossRS for a new target language."""
    with get_session(language) as session:
        if session.query(Sentence).limit(1).first():
            typer_raise(f'CrossRS is already initialized for language "{language}".')
        words_by_sentence, word_frequencies = process_corpus(corpus)
        if not word_frequencies:
            typer_raise('No valid sentences found in the corpus.')
        ids_by_word = add_words(session, word_frequencies)
        add_sentences(session, words_by_sentence, ids_by_word)
        session.add(Metadata(id=1, total_rounds=0))
        print('Committing changes to the database...')
        session.commit()
        print('Optimizing the database...')
        session.execute(text('vacuum'))
    print(f'Initialized CrossRS for language "{language}" '
          f'with {len(words_by_sentence)} unique sentences '
          f'and {len(word_frequencies)} unique words.')



from __future__ import annotations
from datetime import timedelta

from crossrs.db import Session
from crossrs.db.models import Metadata, Word, Sentence
from crossrs.diff.tokenizer import tokenize
from crossrs.utils.time import get_timestamp

__all__ = [
    'update_sentence',
    'undo_learned',
]

INTERVALS = [
    timedelta(days=1, hours=-4),   # 20 hours
    timedelta(days=7, hours=-4),   # 164 hours
    timedelta(days=30, hours=-4),  # ~716 hours
]


def extract_words_from_text(text: str) -> set[str]:
    """Extract all words from a text."""
    return {t.normalized for t in tokenize(text)}


def mark_learned(sentence: Sentence, user_translation: str | None, session: Session) -> None:
    """Mark a sentence as learned and update learnedness for relevant words."""
    sentence.status = 2

    # Collect words from the original sentence
    sentence_word_ids = {word.id for word in sentence.words}

    # Collect words from the user's last translation
    translation_words = extract_words_from_text(user_translation) if user_translation else set()

    # Find word IDs that match the translation word strings
    if translation_words:
        matching = session.query(Word).filter(Word.word.in_(translation_words)).all()
        translation_word_ids = {w.id for w in matching}
    else:
        translation_word_ids = set()

    # Union of both sets
    all_word_ids = sentence_word_ids | translation_word_ids

    # Increment learnedness
    for word_id in all_word_ids:
        session.query(Word).filter(Word.id == word_id).update(
            {Word.learnedness: Word.learnedness + 1},
        )


def undo_learned(sentence: Sentence, session: Session) -> None:
    """Revert learnedness increments if a learned sentence is removed."""
    if sentence.status != 2:
        return

    sentence_word_ids = {word.id for word in sentence.words}

    for word_id in sentence_word_ids:
        session.query(Word).filter(Word.id == word_id, Word.learnedness > 0).update(
            {Word.learnedness: Word.learnedness - 1},
        )


def update_sentence(sentence: Sentence, is_correct: bool, is_first_attempt: bool,
                    target_word: Word | None, user_translation: str,
                    session: Session) -> None:
    """Update sentence state based on evaluation result."""
    now = get_timestamp()
    sentence.rounds += 1
    session.query(Metadata).filter(Metadata.id == 1).update(
        {Metadata.total_rounds: Metadata.total_rounds + 1},
    )

    if is_first_attempt:
        # First time seeing this sentence
        if target_word is not None:
            sentence.target_word_id = target_word.id
        if is_correct:
            # Schedule to the last review stage (30 days - 4 hours)
            sentence.status = 1
            sentence.review_stage = len(INTERVALS) - 1
            sentence.due = now + int(INTERVALS[-1].total_seconds())
        else:
            # Enter the queue at the first stage
            sentence.status = 1
            sentence.review_stage = 0
            sentence.due = now + int(INTERVALS[0].total_seconds())
    else:
        # Review of an in-queue sentence
        if is_correct:
            next_stage = sentence.review_stage + 1
            if next_stage >= len(INTERVALS):
                # All review stages passed — learned
                mark_learned(sentence, user_translation, session)
            else:
                sentence.review_stage = next_stage
                sentence.due = now + int(INTERVALS[next_stage].total_seconds())
        else:
            # Failed review — reset schedule
            sentence.review_stage = 0
            sentence.due = now + int(INTERVALS[0].total_seconds())



from __future__ import annotations
from typing import Annotated

from dataclasses import dataclass
from rich.console import Console
from rich.text import Text
from sqlalchemy import func, case
from typer import Argument, Option

from crossrs.app.app import app
from crossrs.db import get_session, Session
from crossrs.db.models import Metadata, Word, Sentence

__all__ = [
    'stats',
]

DEFAULT_THRESHOLD = 3


@dataclass
class WordStatsItem:
    learned: int = 0
    total: int = 0
    learned_occurrences: int = 0
    total_occurrences: int = 0


@dataclass
class SentenceStatsData:
    learned: int = 0
    in_queue: int = 0
    total: int = 0
    total_rounds: int = 0
    targeted_words: int = 0


def compute_word_stats(session: Session, threshold: int) -> WordStatsItem:
    """Compute word statistics using SQL aggregation."""
    row = session.query(
        func.count(Word.id),
        func.sum(Word.occurrences),
        func.sum(case((Word.learnedness >= threshold, 1), else_=0)),
        func.sum(case((Word.learnedness >= threshold, Word.occurrences), else_=0)),
    ).one()

    return WordStatsItem(
        learned=int(row[2] or 0),
        total=int(row[0] or 0),
        learned_occurrences=int(row[3] or 0),
        total_occurrences=int(row[1] or 0),
    )


def compute_sentence_stats(session: Session, threshold: int) -> SentenceStatsData:
    """Compute sentence statistics using SQL aggregation."""
    row = session.query(
        func.count(Sentence.id),
        func.sum(case((Sentence.status == 2, 1), else_=0)),
        func.sum(case((Sentence.status == 1, 1), else_=0)),
    ).one()

    meta = session.get(Metadata, 1)
    total_rounds = meta.total_rounds if meta else 0

    targeted_words = session.query(
        func.count(func.distinct(Sentence.target_word_id)),
    ).filter(
        Sentence.status == 1,
        Sentence.target_word_id.is_not(None),
    ).join(Word, Sentence.target_word_id == Word.id).filter(
        Word.learnedness < threshold,
    ).scalar()

    return SentenceStatsData(
        total=int(row[0]),
        total_rounds=total_rounds,
        learned=int(row[1] or 0),
        in_queue=int(row[2] or 0),
        targeted_words=int(targeted_words or 0),
    )


def format_section_title(title: str) -> Text:
    return Text(title, style='bold underline')


def format_stats_label(title: str) -> Text:
    return Text(f'{title}:', style='bold')


@app.command()
def stats(
        language: Annotated[
            str,
            Argument(help='Target language code to show statistics for.'),
        ],
        threshold: Annotated[
            int,
            Option(
                '--threshold', '-t',
                help='Learnedness threshold for words to be considered fully learned.',
            ),
        ] = DEFAULT_THRESHOLD,
) -> None:
    """Display study statistics for the given language."""
    with get_session(language) as session:
        word_stats = compute_word_stats(session, threshold)
        sentence_stats = compute_sentence_stats(session, threshold)

    console = Console(highlight=False)

    # Sentence statistics
    console.print(format_section_title('Sentence Statistics'))
    console.print(
        format_stats_label('Sentences'),
        Text(f'{sentence_stats.learned} learned + {sentence_stats.in_queue} in queue '
             f'/ {sentence_stats.total} total'),
    )
    console.print()

    # Word statistics
    console.print(format_section_title('Word Statistics'))
    text = Text(f'{word_stats.learned} learned / {word_stats.total} total')
    if word_stats.total_occurrences > 0:
        coverage = word_stats.learned_occurrences / word_stats.total_occurrences
        text.append(f' ')
        text.append(Text(f'(coverage: {coverage:.1%})', style='dim'))
    console.print(format_stats_label('Words'), text)
    if sentence_stats.targeted_words > 0:
        console.print(
            format_stats_label('Targeted'),
            Text(f'{sentence_stats.targeted_words} unlearned words in queue'),
        )
    console.print()

    # Total rounds
    console.print(format_stats_label('Total rounds'), sentence_stats.total_rounds)



from __future__ import annotations
from dataclasses import dataclass
from difflib import SequenceMatcher

from .tokenizer import tokenize

__all__ = [
    'compute_diff',
    'DiffSegment',
]


@dataclass
class DiffSegment:
    old_start: int
    old_end: int
    old_text: str | None
    new_start: int
    new_end: int
    new_text: str | None


def compute_diff(old: str, new: str) -> list[DiffSegment]:
    """Compute the diff between two texts, returning segments where they differ."""
    segments: list[DiffSegment] = []
    old_tokens = tokenize(old)
    new_tokens = tokenize(new)
    for (tag, i1, i2, j1, j2) in SequenceMatcher(
        a=old_tokens, b=new_tokens, autojunk=False,
    ).get_opcodes():
        if tag != 'equal':
            segments.append(DiffSegment(
                old_start=old_tokens[i1].start if i2 > i1 else (old_tokens[i1 - 1].end if i1 > 0 else 0),
                old_end=old_tokens[i2 - 1].end if i2 > i1 else (old_tokens[i1 - 1].end if i1 > 0 else 0),
                old_text=old[old_tokens[i1].start:old_tokens[i2 - 1].end] if i2 > i1 else None,
                new_start=new_tokens[j1].start if j2 > j1 else (new_tokens[j1 - 1].end if j1 > 0 else 0),
                new_end=new_tokens[j2 - 1].end if j2 > j1 else (new_tokens[j1 - 1].end if j1 > 0 else 0),
                new_text=new[new_tokens[j1].start:new_tokens[j2 - 1].end] if j2 > j1 else None,
            ))
    return segments



from __future__ import annotations
from dataclasses import dataclass
from unicodedata import normalize

import regex

__all__ = [
    'Token',
    'tokenize',
    'normalize_token',
]

PATTERN = regex.compile(r'[\p{L}\p{M}\p{N}]+')


@dataclass(frozen=True)
class Token:
    start: int
    end: int
    normalized: str

    def __eq__(self, other: Token) -> bool:
        return other.normalized == self.normalized

    def __hash__(self) -> int:
        return hash(self.normalized)


def normalize_token(token: str) -> str:
    return normalize('NFC', token.lower())


def tokenize(text: str) -> list[Token]:
    return [
        Token(
            start=match.start(),
            end=match.end(),
            normalized=normalize_token(match.group()),
        )
        for match in PATTERN.finditer(text)
    ]



import os
import threading
from typing import Annotated

from rich.console import Console
from rich.live import Live
from rich.markdown import Markdown
from rich.text import Text
from typer import Argument, Option

from crossrs.app.app import app
from crossrs.db import get_session
from crossrs.db.models import Sentence, Word
from crossrs.diff import compute_diff
from crossrs.utils.console import clear_previous
from .chooser import choose_next, get_round_counts
from .evaluator import evaluate, translate_to_source
from .explainer import explain
from .interaction import ask, ExtraOption
from .updater import update_sentence, undo_learned

__all__ = [
    'study',
]

DEFAULT_THRESHOLD = 3

SECTION_DELIMITER = Text('-' * 3, style='dim')

CORRECT_MARK = '✅'
INCORRECT_MARK = '❌'

EVALUATION_IN_PROGRESS = Text('Evaluating translation...', style='dim')
PREPARING_NEXT = Text('Preparing next sentence...', style='dim')
ENTER_TO_CONTINUE = Text('Press Enter to continue...', style='dim')

QUIT_TITLE = 'Quit'
EXPLAIN_TITLE = 'Explain'
REMOVE_TITLE = 'Remove sentence'
MARK_CORRECT_TITLE = 'Mark correct'

CONSOLE = Console(highlight=False)

QUIT_OPTION = [ExtraOption(QUIT_TITLE, lambda: exit(0))]


def build_explanation_option(source_translation: str,
                             user_translation: str, target_lang: str,
                             source_lang: str, model: str,
                             api_key: str) -> list[ExtraOption]:
    """Build an option to explain why the user's translation was incorrect."""
    def do_explain():
        chunks: list[str] = []
        with CONSOLE.screen():
            with Live(console=CONSOLE, transient=True) as live:
                for chunk in explain(source_translation, user_translation,
                                     target_lang, source_lang, model, api_key):
                    chunks.append(chunk)
                    live.update(Markdown(''.join(chunks)))
        os.environ['LESS'] = '-R'
        with CONSOLE.pager(styles=True):
            CONSOLE.print(Markdown(''.join(chunks)))

    return [ExtraOption(EXPLAIN_TITLE, do_explain)]


def format_user_diff(user_text: str, segments: list) -> Text:
    """Format the user's text with changed/deleted parts highlighted in red."""
    result = Text()
    last_old_end = 0
    for seg in segments:
        if seg.old_text is not None:
            result.append(user_text[last_old_end:seg.old_start])
            result.append(seg.old_text, style='red bold')
            last_old_end = seg.old_end
        elif seg.new_text is not None:
            result.append(user_text[last_old_end:seg.old_start])
            last_old_end = seg.old_start
    result.append(user_text[last_old_end:])
    return result


def format_corrected_diff(corrected_text: str, segments: list) -> Text:
    """Format the corrected text with changed/inserted parts highlighted in green."""
    result = Text()
    last_new_end = 0
    for seg in segments:
        if seg.new_text is not None:
            result.append(corrected_text[last_new_end:seg.new_start])
            result.append(seg.new_text, style='green bold')
            last_new_end = seg.new_end
        elif seg.old_text is not None:
            result.append(corrected_text[last_new_end:seg.new_start])
            last_new_end = seg.new_start
    result.append(corrected_text[last_new_end:])
    return result


@app.command()
def study(
        language: Annotated[
            str,
            Argument(help='Target language code previously initialized with `crossrs init`.'),
        ],
        source: Annotated[
            str,
            Argument(help='Source language code to translate into (e.g., "en").'),
        ],
        threshold: Annotated[
            int,
            Option(
                '--threshold', '-t',
                help='Learnedness threshold for words to be considered fully learned.',
            ),
        ] = DEFAULT_THRESHOLD,
        model: Annotated[
            str,
            Option(
                '--model',
                envvar='CROSSRS_MODEL',
                help='GPT model for translation and evaluation.',
            ),
        ] = 'gpt-4.1-mini',
        api_key: Annotated[
            str,
            Option(
                '--api-key',
                envvar='CROSSRS_API_KEY',
                help='OpenAI API key.',
            ),
        ] = '',
) -> None:
    """Launch an interactive study session."""
    if not api_key:
        from crossrs.utils.typer import typer_raise
        typer_raise('API key is required. Set --api-key or CROSSRS_API_KEY environment variable.')

    with get_session(language) as session:
        sentence: Sentence
        target_word: Word | None
        source_translation: str

        # Background preparation results (IDs to be merged into main session)
        _bg_sentence_id: int
        _bg_target_word_id: int | None
        _bg_error: BaseException | None = None

        def prepare_next() -> None:
            """Choose the next sentence and translate it in a separate session."""
            nonlocal _bg_sentence_id, _bg_target_word_id, source_translation, _bg_error
            _bg_error = None
            try:
                with get_session(language) as bg_session:
                    bg_sentence, bg_target_word, _ = choose_next(bg_session, threshold)
                    _bg_sentence_id = bg_sentence.id
                    _bg_target_word_id = bg_target_word.id if bg_target_word else None
                    source_translation = translate_to_source(
                        bg_sentence, language, source, model, api_key,
                    )
            except BaseException as e:
                _bg_error = e

        def load_prepared() -> None:
            """Load the background-prepared sentence into the main session."""
            nonlocal sentence, target_word
            if _bg_error is not None:
                raise _bg_error
            session.expire_all()
            sentence = session.get(Sentence, _bg_sentence_id)
            target_word = session.get(Word, _bg_target_word_id) if _bg_target_word_id else None

        CONSOLE.print(Text('Preparing first sentence...', style='dim'))
        prepare_next()
        load_prepared()
        clear_previous()

        bg_ready: threading.Event | None = None

        while True:
            # Wait for background preparation if it hasn't finished yet
            if bg_ready is not None and not bg_ready.is_set():
                CONSOLE.print(PREPARING_NEXT)
                bg_ready.wait()
                clear_previous()
            if bg_ready is not None:
                try:
                    load_prepared()
                except Exception as e:
                    CONSOLE.print(Text(f'Error: {e}', style='red'))
                    CONSOLE.print(Text('Retrying...', style='dim'))
                    prepare_next()
                    load_prepared()
                    clear_previous(2)

            is_first_attempt = sentence.status == 0

            # 1) Show round status (always fresh)
            counts = get_round_counts(session)
            status_parts: list[Text] = []
            if counts.due > 0:
                status_parts.append(Text(f'{counts.due} due', style='yellow bold'))
            if counts.in_queue > 0:
                status_parts.append(Text(f'{counts.in_queue} in queue', style='dim'))
            if counts.learned > 0:
                status_parts.append(Text(f'{counts.learned} learned', style='green'))
            if status_parts:
                line = status_parts[0]
                for part in status_parts[1:]:
                    line = line + Text(' · ') + part
                CONSOLE.print(line)

            # 2) Show source translation and get user input
            formatted_prompt = Text(source_translation, style='bold')
            user_translation = ask(CONSOLE, formatted_prompt, QUIT_OPTION)
            clear_previous(2)
            CONSOLE.print(formatted_prompt)
            CONSOLE.print(user_translation)

            # 3) Evaluate
            CONSOLE.print(EVALUATION_IN_PROGRESS)
            try:
                evaluation = evaluate(
                    sentence, source_translation, user_translation,
                    language, source, model, api_key,
                )
            except Exception as e:
                clear_previous()
                CONSOLE.print(Text(f'Error: {e}', style='red'))
                CONSOLE.print(Text('Press Enter to retry...', style='dim'))
                input()
                clear_previous(4)
                continue
            clear_previous()

            # 4) Show result
            CONSOLE.print(SECTION_DELIMITER)
            if evaluation.is_correct:
                CONSOLE.print(Text(f'{CORRECT_MARK} ', style='') + Text(user_translation))
            else:
                CONSOLE.print(Text(f'{INCORRECT_MARK} '))
                if evaluation.corrected_translation:
                    segments = compute_diff(user_translation, evaluation.corrected_translation)
                    if segments:
                        CONSOLE.print(format_user_diff(user_translation, segments))
                        CONSOLE.print(format_corrected_diff(evaluation.corrected_translation, segments))
                    else:
                        CONSOLE.print(Text(evaluation.corrected_translation, style='bold'))
                CONSOLE.print(
                    Text('Original: ', style='dim')
                    + Text(sentence.sentence, style='bold italic'),
                )
            CONSOLE.print(SECTION_DELIMITER)

            # 5) Update (save state first so we can undo if user marks correct)
            saved_status = sentence.status
            saved_stage = sentence.review_stage
            saved_due = sentence.due
            saved_rounds = sentence.rounds
            update_sentence(sentence, evaluation.is_correct, is_first_attempt,
                            target_word, user_translation, session)
            session.commit()

            # Save reference to the sentence shown this round
            shown_sentence = sentence

            # 6) Start background preparation, then show continue prompt
            event = threading.Event()
            bg_ready = event
            threading.Thread(
                target=lambda e=event: (prepare_next(), e.set()),
                daemon=True,
            ).start()

            removed = False
            marked_correct = False

            def do_remove() -> None:
                nonlocal removed
                removed = True

            def do_mark_correct() -> None:
                nonlocal marked_correct
                marked_correct = True

            explain_option = (
                build_explanation_option(source_translation, user_translation,
                                        language, source, model, api_key)
                if not evaluation.is_correct else []
            )
            mark_correct_option = (
                [ExtraOption(MARK_CORRECT_TITLE, do_mark_correct, key='MC', returns=True)]
                if not evaluation.is_correct else []
            )
            remove_option = [ExtraOption(REMOVE_TITLE, do_remove, key='RS', returns=True)]
            ask(CONSOLE, ENTER_TO_CONTINUE,
                explain_option + mark_correct_option + remove_option + QUIT_OPTION)

            if marked_correct:
                # Undo incorrect update and redo as correct
                shown_sentence.status = saved_status
                shown_sentence.review_stage = saved_stage
                shown_sentence.due = saved_due
                shown_sentence.rounds = saved_rounds
                update_sentence(shown_sentence, True, is_first_attempt,
                                target_word, user_translation, session)
                session.commit()
                bg_ready.wait()
                clear_previous(2)
                CONSOLE.print(Text(f'{CORRECT_MARK}  Marked as correct.', style='dim'))
                continue

            if removed:
                undo_learned(shown_sentence, session)
                session.delete(shown_sentence)
                session.commit()
                bg_ready.wait()
                clear_previous(2)
                CONSOLE.print(Text('Sentence removed.', style='dim'))
                continue

            clear_previous(2)
            CONSOLE.print(SECTION_DELIMITER)



from __future__ import annotations

from dataclasses import dataclass

from sqlalchemy import case, func

from crossrs.db import Session
from crossrs.db.models import Word, Sentence, SentenceWord
from crossrs.utils.time import get_timestamp
from crossrs.utils.typer import typer_raise

__all__ = [
    'choose_next',
]

MAX_CANDIDATE_SENTENCES = 10000


@dataclass
class RoundCounts:
    due: int = 0
    in_queue: int = 0
    learned: int = 0


def get_round_counts(session: Session) -> RoundCounts:
    """Return due, in-queue, and learned sentence counts in a single query."""
    now = get_timestamp()
    row = session.query(
        func.sum(case((
            (Sentence.status == 1) & Sentence.due.is_not(None) & (Sentence.due <= now), 1,
        ), else_=0)),
        func.sum(case((Sentence.status == 1, 1), else_=0)),
        func.sum(case((Sentence.status == 2, 1), else_=0)),
    ).one()
    return RoundCounts(
        due=int(row[0] or 0),
        in_queue=int(row[1] or 0),
        learned=int(row[2] or 0),
    )


def choose_due_sentence(session: Session) -> Sentence | None:
    """Return the sentence with the earliest due timestamp, or None."""
    return session.query(Sentence).filter(
        Sentence.status == 1,
        Sentence.due.is_not(None),
        Sentence.due <= get_timestamp(),
    ).order_by(Sentence.due).limit(1).first()


def get_next_due_time(session: Session) -> int | None:
    """Return the earliest future due timestamp, or None if no sentences are in the queue."""
    result = session.query(func.min(Sentence.due)).filter(
        Sentence.status == 1,
        Sentence.due.is_not(None),
    ).scalar()
    return result


def get_queue_target_word_ids(session: Session) -> set[int]:
    """Return the set of word IDs that are targets of sentences currently in the queue."""
    rows = session.query(Sentence.target_word_id).filter(
        Sentence.status == 1,
        Sentence.target_word_id.is_not(None),
    ).all()
    return {row[0] for row in rows}


def choose_target_word(session: Session, threshold: int) -> Word | None:
    """
    Choose the next target word using a single SQL query with EXISTS subquery:
    1. learnedness < threshold
    2. Not already a target of any in-queue sentence
    3. Has at least one unseen sentence
    4. Highest learnedness first
    5. Highest occurrences (lowest id) among ties
    """
    queue_targets = get_queue_target_word_ids(session)

    # Correlated EXISTS subquery: does this word appear in any unseen sentence?
    unseen_exists = session.query(SentenceWord.sentence_id).join(
        Sentence, SentenceWord.sentence_id == Sentence.id,
    ).filter(
        SentenceWord.word_id == Word.id,
        Sentence.status == 0,
    ).exists()

    query = session.query(Word).filter(
        Word.learnedness < threshold,
        unseen_exists,
    )
    if queue_targets:
        query = query.filter(~Word.id.in_(queue_targets))

    return query.order_by(Word.learnedness.desc(), Word.id).limit(1).first()


def choose_sentence_for_word(session: Session, word: Word, threshold: int) -> Sentence:
    """
    Choose a sentence for the given word. Samples MAX_CANDIDATE_SENTENCES random unseen
    sentences, then ranks them entirely in SQL:
    1. Fewest unseen words (learnedness == 0)
    2. Most seen-but-not-learned words (0 < learnedness < threshold)
    3. Smallest sum of word ids (frequency rankings)
    """
    candidates_sq = (
        session.query(SentenceWord.sentence_id)
        .join(Sentence, SentenceWord.sentence_id == Sentence.id)
        .filter(
            SentenceWord.word_id == word.id,
            Sentence.status == 0,
        )
        .order_by(SentenceWord.random_key)
        .limit(MAX_CANDIDATE_SENTENCES)
        .subquery()
    )

    best = (
        session.query(candidates_sq.c.sentence_id)
        .join(SentenceWord, candidates_sq.c.sentence_id == SentenceWord.sentence_id)
        .join(Word, SentenceWord.word_id == Word.id)
        .group_by(candidates_sq.c.sentence_id)
        .order_by(
            func.sum(case((Word.learnedness == 0, 1), else_=0)),
            func.sum(case((Word.learnedness.between(1, threshold - 1), 1), else_=0)).desc(),
            func.sum(Word.id),
        )
        .limit(1)
        .first()
    )

    return session.get(Sentence, best[0])


def choose_next(session: Session, threshold: int) -> tuple[Sentence, Word | None, RoundCounts]:
    """
    Return the next sentence to study, the target word (None for review sentences),
    and the current round counts.
    """
    counts = get_round_counts(session)

    # Priority 1: due review sentence
    if due_sentence := choose_due_sentence(session):
        return due_sentence, None, counts

    # Priority 2: new sentence via word algorithm
    if target_word := choose_target_word(session, threshold):
        sentence = choose_sentence_for_word(session, target_word, threshold)
        return sentence, target_word, counts

    # Priority 3: future due sentence exists — inform user
    if next_due := get_next_due_time(session):
        now = get_timestamp()
        wait_seconds = max(0, next_due - now)
        hours = wait_seconds // 3600
        minutes = (wait_seconds % 3600) // 60
        if hours > 0:
            time_str = f'{hours}h {minutes}m'
        else:
            time_str = f'{minutes}m'
        typer_raise(f'No sentences to review right now. Next review due in {time_str}.')

    # Nothing left
    typer_raise('🎉 Congratulations! You have learned all the words in the corpus!')



from __future__ import annotations

from sqlalchemy import ForeignKey, Index, text
from sqlalchemy.orm import Mapped, mapped_column, relationship
from .base import Base

__all__ = [
    'EvaluationCache',
    'Metadata',
    'Word',
    'Sentence',
    'SentenceWord',
    'TranslationCache',
]


class Metadata(Base):
    __tablename__ = 'metadata'

    id: Mapped[int] = mapped_column(primary_key=True, default=1)
    total_rounds: Mapped[int] = mapped_column(default=0)


class Word(Base):
    __tablename__ = 'words'

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    word: Mapped[str]
    occurrences: Mapped[int]
    learnedness: Mapped[int] = mapped_column(default=0)

    sentences: Mapped[list[Sentence]] = relationship(
        'Sentence',
        secondary='sentence_words',
        order_by=lambda: SentenceWord.random_key,
        back_populates='words',
        passive_deletes=True,
    )


Index('ix_words_learnedness_id', Word.learnedness, Word.id)


class Sentence(Base):
    __tablename__ = 'sentences'

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    sentence: Mapped[str]
    status: Mapped[int] = mapped_column(default=0)  # 0=unseen, 1=in_queue, 2=learned
    target_word_id: Mapped[int | None] = mapped_column(
        ForeignKey('words.id', ondelete='SET NULL', onupdate='CASCADE'),
        default=None,
    )
    due: Mapped[int | None] = mapped_column(default=None)  # Unix timestamp
    review_stage: Mapped[int] = mapped_column(default=0)
    rounds: Mapped[int] = mapped_column(default=0)

    words: Mapped[list[Word]] = relationship(
        'Word',
        secondary='sentence_words',
        order_by=Word.id,
        back_populates='sentences',
        passive_deletes=True,
    )


Index('ix_sentences_due_in_queue', Sentence.due, sqlite_where=Sentence.status == 1)
Index('ix_sentences_target_word_in_queue', Sentence.target_word_id,
      sqlite_where=(Sentence.status == 1) & (Sentence.target_word_id.is_not(None)))
Index('ix_sentences_status', Sentence.status)


class SentenceWord(Base):
    __tablename__ = 'sentence_words'

    sentence_id: Mapped[int] = mapped_column(
        ForeignKey('sentences.id', ondelete='CASCADE', onupdate='CASCADE'),
        primary_key=True,
    )
    word_id: Mapped[int] = mapped_column(
        ForeignKey('words.id', ondelete='CASCADE', onupdate='CASCADE'),
        primary_key=True,
    )
    random_key: Mapped[int] = mapped_column(server_default=text('(abs(random()) % 16384)'))

    __table_args__ = (
        Index('ix_sentence_words_word_id_random_key', 'word_id', 'random_key'),
        {'sqlite_with_rowid': False},
    )


class TranslationCache(Base):
    __tablename__ = 'translation_cache'

    sentence_id: Mapped[int] = mapped_column(
        ForeignKey('sentences.id', ondelete='CASCADE', onupdate='CASCADE'),
        primary_key=True,
    )
    source_lang: Mapped[str] = mapped_column(primary_key=True)
    translation: Mapped[str]
    model: Mapped[str]


class EvaluationCache(Base):
    __tablename__ = 'evaluation_cache'

    sentence_id: Mapped[int] = mapped_column(
        ForeignKey('sentences.id', ondelete='CASCADE', onupdate='CASCADE'),
        primary_key=True,
    )
    model: Mapped[str] = mapped_column(primary_key=True)
    translation: Mapped[str] = mapped_column(primary_key=True)
    evaluation: Mapped[str]



from __future__ import annotations
from dataclasses import dataclass

from openai import OpenAI
from pydantic import BaseModel

from crossrs.db import get_session
from crossrs.db.models import EvaluationCache, Sentence, TranslationCache
from crossrs.utils.openai import api_call_with_retries
from crossrs.utils.strings import normalize

__all__ = [
    'evaluate',
    'translate_to_source',
    'Evaluation',
]

TEMPERATURE = 0.0


@dataclass
class Evaluation:
    is_correct: bool
    corrected_translation: str | None


class TranslationOutput(BaseModel):
    translation: str


class EvaluationOutput(BaseModel):
    is_correct: bool
    corrected_translation: str | None


def generate_translation_prompt(target_lang: str, source_lang: str, sentence: str) -> str:
    """Generate a prompt to translate a sentence from target language to source language."""
    return '\n'.join([
        f'Translate the following sentence from the language with code `{target_lang}` '
        f'into the language with code `{source_lang}`. '
        f'Provide a natural, accurate translation.',
        f'',
        f'`{sentence}`',
    ])


def generate_evaluation_prompt(
        target_lang: str, source_lang: str,
        source_translation: str, user_translation: str,
) -> str:
    """Generate a prompt to evaluate the user's translation back into the target language."""
    return '\n'.join([
        f'A learner is practicing the language with code `{target_lang}` by translating '
        f'sentences from `{source_lang}` back into `{target_lang}`.',
        f'',
        f'The sentence in `{source_lang}`:',
        f'`{source_translation}`',
        f'',
        f"The learner's translation into `{target_lang}`:",
        f'`{user_translation}`',
        f'',
        f"Determine whether the learner's translation is fully grammatical, natural, "
        f'and conveys the same meaning as the `{source_lang}` sentence.',
        f'',
        f"If the learner's translation is correct (or differs only in trivial ways like "
        f'punctuation), set `is_correct` to true and `corrected_translation` to null.',
        f'',
        f'If the translation has any grammatical, lexical, or meaning issues, set '
        f'`is_correct` to false and provide a `corrected_translation` that is a '
        f"**minimally edited** version of the learner's attempt that makes it fully "
        f'grammatical, natural, and meaning-preserving. Change as few words as possible.',
    ])


def request_translation(sentence: str, target_lang: str, source_lang: str, model: str, api_key: str) -> str:
    """Request translation of a sentence from target to source language."""
    return api_call_with_retries(lambda: OpenAI(api_key=api_key).responses.parse(
        model=model,
        temperature=TEMPERATURE,
        text_format=TranslationOutput,
        input=generate_translation_prompt(target_lang, source_lang, sentence),
    ).output_parsed.translation)


def translate_to_source(sentence: Sentence, target_lang: str, source_lang: str,
                        model: str, api_key: str) -> str:
    """Translate a sentence to the source language, using cache if available."""
    with get_session(target_lang) as session:
        if cached := session.query(TranslationCache).filter(
            TranslationCache.sentence_id == sentence.id,
            TranslationCache.source_lang == source_lang,
        ).limit(1).first():
            return cached.translation
        else:
            translation = request_translation(sentence.sentence, target_lang, source_lang,
                                              model, api_key)
            session.add(TranslationCache(
                sentence_id=sentence.id,
                source_lang=source_lang,
                translation=translation,
                model=model,
            ))
            return translation


def request_evaluation(source_translation: str, user_translation: str,
                       target_lang: str, source_lang: str, model: str,
                       api_key: str) -> EvaluationOutput:
    """Request LLM evaluation of the user's translation."""
    return api_call_with_retries(lambda: OpenAI(api_key=api_key).responses.parse(
        model=model,
        temperature=TEMPERATURE,
        text_format=EvaluationOutput,
        input=generate_evaluation_prompt(target_lang, source_lang,
                                         source_translation, user_translation),
    ).output_parsed)


def cached_evaluation(sentence: Sentence, source_translation: str, user_translation: str,
                      target_lang: str, source_lang: str, model: str,
                      api_key: str) -> EvaluationOutput:
    """Request evaluation with caching."""
    with get_session(target_lang) as session:
        if cached := session.query(EvaluationCache).filter(
            EvaluationCache.sentence_id == sentence.id,
            EvaluationCache.model == model,
            EvaluationCache.translation == user_translation,
        ).limit(1).first():
            return EvaluationOutput.model_validate_json(cached.evaluation)
        else:
            output = request_evaluation(source_translation,
                                        user_translation, target_lang, source_lang,
                                        model, api_key)
            session.add(EvaluationCache(
                sentence_id=sentence.id,
                model=model,
                translation=user_translation,
                evaluation=output.model_dump_json(),
            ))
            return output


def build_evaluation(output: EvaluationOutput, sentence: Sentence,
                     user_translation: str) -> Evaluation:
    """Build an Evaluation from the LLM output."""
    is_correct = (output.is_correct
                  or normalize(output.corrected_translation or '') == normalize(user_translation)
                  or normalize(user_translation) == normalize(sentence.sentence))
    return Evaluation(
        is_correct=is_correct,
        corrected_translation=None if is_correct else output.corrected_translation,
    )


def evaluate(sentence: Sentence, source_translation: str, user_translation: str,
             target_lang: str, source_lang: str, model: str,
             api_key: str) -> Evaluation:
    """Evaluate the user's translation of a sentence."""
    if normalize(user_translation) == normalize(sentence.sentence):
        return Evaluation(is_correct=True, corrected_translation=None)
    output = cached_evaluation(sentence, source_translation, user_translation,
                               target_lang, source_lang, model, api_key)
    return build_evaluation(output, sentence, user_translation)



from dataclasses import dataclass
from typing import Callable

from prompt_toolkit import PromptSession
from rich.console import Console
from rich.text import Text

from crossrs.utils.console import clear_previous

__all__ = [
    'ask',
    'ExtraOption',
]

OPTION_DELIMITER = ' | '


@dataclass
class ExtraOption:
    title: str
    action: Callable[[], None]
    key: str | None = None  # Custom key; defaults to first character of title
    returns: bool = False  # If True, ask() returns empty string after action


def normalize_key(text: str) -> str:
    return text.lower()


def build_options(extra_options: list[ExtraOption]) -> tuple[list[list[Text]], dict[str, ExtraOption]]:
    """Build the list of option prompts and a mapping from option keys to their actions."""
    prompt = []
    mapping: dict[str, ExtraOption] = {}
    for option in extra_options:
        if not option.title:
            raise ValueError('Option title cannot be empty.')
        key = option.key or option.title[0]
        key_lower = normalize_key(key)
        if key_lower in mapping:
            raise ValueError(f'Duplicate key for options: {key_lower}')
        mapping[key_lower] = option
        # Bold the key characters: if key matches first letters of words, bold those
        if len(key) > 1:
            words = option.title.split()
            parts: list[Text] = []
            for i, word in enumerate(words):
                if i < len(key):
                    parts.append(Text(word[0], style='bold'))
                    parts.append(Text(word[1:]))
                else:
                    parts.append(Text(word))
                if i < len(words) - 1:
                    parts.append(Text(' '))
            prompt.append(parts)
        else:
            prompt.append([Text(option.title[0], style='bold'), Text(option.title[1:])])
    return prompt, mapping


def ask(
        console: Console,
        prompt: Text,
        extra_options: list[ExtraOption],
) -> str:
    """Prompt the user for input or to choose one of the extra options."""
    options_prompt, options_mapping = build_options(extra_options)
    console.print(prompt, end='')
    for option_prompt in options_prompt:
        console.print(Text(OPTION_DELIMITER, style='dim'), end='')
        console.print(*option_prompt, sep='', end='')
    console.print()
    while True:
        response = PromptSession().prompt()
        if option := options_mapping.get(normalize_key(response)):
            clear_previous()
            option.action()
            if option.returns:
                return ''
        else:
            return response.strip()



from openai import OpenAI
from typing import Generator

from crossrs.utils.openai import api_call_with_retries

__all__ = [
    'explain',
]

TEMPERATURE = 0.0


def generate_prompt(target_lang: str, source_lang: str,
                    source_translation: str, user_translation: str) -> str:
    """Generate a prompt instructing the LLM to explain why the user's translation was incorrect."""
    return '\n'.join([
        f'A learner is practicing the language with code `{target_lang}` by translating '
        f'sentences from `{source_lang}` back into `{target_lang}`.',
        f'',
        f'The sentence in `{source_lang}`:',
        f'`{source_translation}`',
        f'',
        f"The learner's translation into `{target_lang}`:",
        f'`{user_translation}`',
        f'',
        f"Explain to the learner why their translation is not fully correct. "
        f"Focus on the specific grammatical, lexical, or meaning issues. "
        f"Be concise and helpful. Do not mention the language codes "
        f"`{target_lang}` or `{source_lang}` in your explanation.",
    ])


def explain(source_translation: str, user_translation: str,
            target_lang: str, source_lang: str, model: str,
            api_key: str) -> Generator[str, None, None]:
    """Stream an explanation for why the user's translation was incorrect."""
    stream = api_call_with_retries(lambda: OpenAI(api_key=api_key).responses.create(
        model=model,
        temperature=TEMPERATURE,
        input=generate_prompt(target_lang, source_lang,
                              source_translation, user_translation),
        stream=True,
    ))
    for event in stream:
        if event.type == 'response.output_text.delta':
            yield event.delta



__all__ = [
    'clear_current_line',
    'clear_previous',
]


def clear_current_line() -> None:
    print('\033[2K', end='\r')


def clear_previous(lines: int = 1) -> None:
    print('\033[F\033[K' * lines, end='\r')



from __future__ import annotations
from typing import Never

import typer

__all__ = [
    'typer_raise',
]


def typer_raise(message: str) -> Never:
    typer.echo(message, err=True)
    raise typer.Exit(code=1)
